# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import logging

import wrapt

from scout_apm.core.context import (
    SCOUT_REQUEST_ATTR,
    TrackedRequest,
    get_current_asyncio_task,
)

try:
    import asyncio
except ImportError:
    asyncio = None

logger = logging.getLogger(__name__)

have_patched_asyncio = False


def wrapped_create_task(wrapped, instance, args, kwargs):
    """Wrapper for ``create_task(coro)`` that propagates the current
    ``TrackedRequest`` to the new ``Task``.
    """
    new_task = wrapped(*args, **kwargs)
    current_task = get_current_asyncio_task()

    tracked_request = (
        getattr(current_task, SCOUT_REQUEST_ATTR, None) or TrackedRequest.instance()
    )

    if tracked_request.is_real_request and not tracked_request.is_ignored():
        # Updates the ``TrackedRequest`` for the given Task.
        # Used to pass the request amongst different tasks.
        setattr(new_task, SCOUT_REQUEST_ATTR, tracked_request)

    return new_task


def ensure_installed():
    global have_patched_asyncio

    logger.debug("Instrumenting asyncio.")

    if asyncio is None:
        logger.debug("Couldn't import asyncio - probably old version of python.")
    else:
        if not have_patched_asyncio:
            try:
                wrapt.wrap_function_wrapper(
                    asyncio.BaseEventLoop, "create_task", wrapped_create_task
                )
            except Exception as exc:
                logger.warning(
                    "Failed to instrument asyncio.create_task: %r",
                    exc,
                    exc_info=exc,
                )
            else:
                have_patched_asyncio = True

    return True



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import wrapt
from flask import current_app
from flask.globals import request

import scout_apm.core
from scout_apm.core.config import scout_config
from scout_apm.core.tracked_request import TrackedRequest
from scout_apm.core.web_requests import werkzeug_track_request_data


class ScoutApm(object):
    def __init__(self, app):
        self.app = app
        self._attempted_install = False
        app.full_dispatch_request = self.wrapped_full_dispatch_request(
            app.full_dispatch_request
        )
        app.preprocess_request = self.wrapped_preprocess_request(app.preprocess_request)

    @wrapt.decorator
    def wrapped_full_dispatch_request(self, wrapped, instance, args, kwargs):
        if not self._attempted_install:
            self.extract_flask_settings()
            installed = scout_apm.core.install()
            self._do_nothing = not installed
            self._attempted_install = True

        if self._do_nothing:
            return wrapped(*args, **kwargs)

        # Pass on routing exceptions (normally 404's)
        if request.routing_exception is not None:
            return wrapped(*args, **kwargs)

        rule = request.url_rule
        view_func = instance.view_functions[rule.endpoint]

        name = view_func.__module__ + "." + view_func.__name__
        operation = "Controller/" + name

        tracked_request = TrackedRequest.instance()
        tracked_request.is_real_request = True
        request._scout_tracked_request = tracked_request

        werkzeug_track_request_data(request, tracked_request)

        with tracked_request.span(
            operation=operation, should_capture_backtrace=False
        ) as span:
            request._scout_view_span = span

            try:
                response = wrapped(*args, **kwargs)
            except Exception as exc:
                tracked_request.tag("error", "true")
                raise exc
            else:
                if 500 <= response.status_code <= 599:
                    tracked_request.tag("error", "true")
                return response

    @wrapt.decorator
    def wrapped_preprocess_request(self, wrapped, instance, args, kwargs):
        tracked_request = getattr(request, "_scout_tracked_request", None)
        if tracked_request is None:
            return wrapped(*args, **kwargs)

        # Unlike middleware in other frameworks, using request preprocessors is
        # less common in Flask, so only add a span if there is any in use
        have_before_request_funcs = (
            None in instance.before_request_funcs
            or request.blueprint in instance.before_request_funcs
        )
        if not have_before_request_funcs:
            return wrapped(*args, **kwargs)

        with tracked_request.span("PreprocessRequest", should_capture_backtrace=False):
            return wrapped(*args, **kwargs)

    def extract_flask_settings(self):
        """
        Copies SCOUT_* settings in the app into Scout's config lookup
        """
        configs = {}
        configs["application_root"] = self.app.instance_path
        for name in current_app.config:
            if name.startswith("SCOUT_"):
                value = current_app.config[name]
                clean_name = name.replace("SCOUT_", "").lower()
                configs[clean_name] = value
        scout_config.set(**configs)



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import itertools
import os
import sys
import sysconfig
import traceback
import warnings

# Maximum non-Scout frames to target retrieving
LIMIT = 50
# How many upper frames from inside Scout to ignore
IGNORED = 1


def filter_frames(frames):
    """Filter the stack trace frames down to non-library code."""
    paths = sysconfig.get_paths()
    library_paths = {paths["purelib"], paths["platlib"]}
    for frame in frames:
        if not any(frame["file"].startswith(exclusion) for exclusion in library_paths):
            yield frame


def module_filepath(module, filepath):
    """Get the filepath relative to the base module."""
    root_module = module.split(".", 1)[0]
    if root_module == module:
        return os.path.basename(filepath)

    module_dir = sys.modules[root_module].__file__.rsplit(os.sep, 2)[0]
    return filepath.split(module_dir, 1)[-1].lstrip(os.sep)


def filepath(frame):
    """Get the filepath for frame."""
    module = frame.f_globals.get("__name__", None)
    filepath = frame.f_code.co_filename

    if filepath.endswith(".pyc"):
        filepath = filepath[:-1]

    if not module:
        return filepath
    return module_filepath(module, filepath)


if sys.version_info >= (3, 5):

    def stacktrace_walker(tb):
        """Iterate over each frame of the stack downards for exceptions."""
        for frame, lineno in traceback.walk_tb(tb):
            name = frame.f_code.co_name
            yield {"file": filepath(frame), "line": lineno, "function": name}

    def backtrace_walker():
        """Iterate over each frame of the stack upwards.

        Taken from python3/traceback.ExtractSummary.extract to support
        iterating over the entire stack, but without creating a large
        data structure.
        """
        start_frame = sys._getframe().f_back
        for frame, lineno in traceback.walk_stack(start_frame):
            name = frame.f_code.co_name
            yield {"file": filepath(frame), "line": lineno, "function": name}


else:

    def stacktrace_walker(tb):
        """Iterate over each frame of the stack downards for exceptions."""
        while tb is not None:
            lineno = tb.tb_lineno
            name = tb.tb_frame.f_code.co_name
            yield {
                "file": filepath(tb.tb_frame),
                "line": lineno,
                "function": name,
            }
            tb = tb.tb_next

    def backtrace_walker():
        """Iterate over each frame of the stack upwards.

        Taken from python2.7/traceback.extract_stack to support iterating
        over the entire stack, but without creating a large data structure.
        """
        try:
            raise ZeroDivisionError
        except ZeroDivisionError:
            # Get the current frame
            frame = sys.exc_info()[2].tb_frame.f_back

        while frame is not None:
            lineno = frame.f_lineno
            name = frame.f_code.co_name
            yield {"file": filepath(frame), "line": lineno, "function": name}
            frame = frame.f_back


def capture_backtrace():
    walker = filter_frames(backtrace_walker())
    return list(itertools.islice(walker, LIMIT))


def capture_stacktrace(tb):
    walker = stacktrace_walker(tb)
    return list(reversed(list(itertools.islice(walker, LIMIT))))


def capture():
    warnings.warn(
        "capture is deprecated, instead use capture_backtrace instead.",
        DeprecationWarning,
        2,
    )
    return capture_backtrace()



# coding=utf-8
from __future__ import absolute_import, division, print_function, unicode_literals

import cherrypy
from cherrypy.lib.encoding import ResponseEncoder
from cherrypy.process import plugins

import scout_apm.core
from scout_apm.compat import parse_qsl
from scout_apm.core.config import scout_config
from scout_apm.core.tracked_request import TrackedRequest
from scout_apm.core.web_requests import (
    create_filtered_path,
    ignore_path,
    track_request_queue_time,
)


class ScoutPlugin(plugins.SimplePlugin):
    def __init__(self, bus):
        super(ScoutPlugin, self).__init__(bus)
        installed = scout_apm.core.install()
        self._do_nothing = not installed

    def before_request(self):
        if self._do_nothing:
            return
        request = cherrypy.request
        tracked_request = TrackedRequest.instance()
        tracked_request.is_real_request = True
        request._scout_tracked_request = tracked_request

        # Can't name operation until after request, when routing has been done
        request._scout_controller_span = tracked_request.start_span(
            "Controller/Unknown"
        )

    def after_request(self):
        if self._do_nothing:
            return
        request = cherrypy.request
        tracked_request = getattr(request, "_scout_tracked_request", None)
        if tracked_request is None:
            return

        # Rename controller span now routing has been done
        operation_name = get_operation_name(request)
        if operation_name is not None:
            request._scout_controller_span.operation = operation_name

        # Grab general request data now it has been parsed
        path = request.path_info
        # Parse params ourselves because we want only GET params but CherryPy
        # parses POST params (nearly always sensitive) into the same dict.
        params = parse_qsl(request.query_string)
        tracked_request.tag("path", create_filtered_path(path, params))
        if ignore_path(path):
            tracked_request.tag("ignore_transaction", True)

        if scout_config.value("collect_remote_ip"):
            # Determine a remote IP to associate with the request. The value is
            # spoofable by the requester so this is not suitable to use in any
            # security sensitive context.
            user_ip = (
                request.headers.get("x-forwarded-for", "").split(",")[0]
                or request.headers.get("client-ip", "").split(",")[0]
                or (request.remote.ip or None)
            )
            tracked_request.tag("user_ip", user_ip)

        queue_time = request.headers.get("x-queue-start", "") or request.headers.get(
            "x-request-start", ""
        )
        track_request_queue_time(queue_time, tracked_request)

        response = cherrypy.response
        status = response.status
        if isinstance(status, int):
            status_int = status
        else:
            status_first = status.split(" ", 1)[0]
            try:
                status_int = int(status_first)
            except ValueError:
                # Assume OK
                status_int = 200
        if 500 <= status_int <= 599:
            tracked_request.tag("error", "true")
        elif status_int == 404:
            tracked_request.is_real_request = False

        tracked_request.stop_span()


def get_operation_name(request):
    handler = request.handler
    if handler is None:
        return None

    if isinstance(handler, ResponseEncoder):
        real_handler = handler.oldhandler
    else:
        real_handler = handler

    # Unwrap HandlerWrapperTool classes
    while hasattr(real_handler, "callable"):
        real_handler = real_handler.callable

    # Looks like it's from HandlerTool
    if getattr(real_handler, "__name__", "") == "handle_func":
        try:
            wrapped_tool = real_handler.__closure__[2].cell_contents.callable
        except (AttributeError, IndexError):
            pass
        else:
            try:
                return "Controller/{}".format(wrapped_tool.__name__)
            except AttributeError:
                pass

    # Not a method? Not from an exposed view then
    if not hasattr(real_handler, "__self__"):
        return None

    real_handler_cls = real_handler.__self__.__class__
    return "Controller/{}.{}.{}".format(
        real_handler_cls.__module__,
        real_handler_cls.__name__,
        real_handler.__name__,
    )
